#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import shutil

from splunk_add_on_ucc_framework import global_config as global_config_lib
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    alert_actions_conf_gen,
    normalize,
)
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    alert_actions_html_gen,
)
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    alert_actions_py_gen,
)

logger = logging.getLogger("ucc_gen")


def generate_alerts(
    global_config: global_config_lib.GlobalConfig,
    addon_name: str,
    internal_source_dir: str,
    output_dir: str,
) -> None:
    envs = normalize.normalize(
        global_config.alerts,
        global_config.namespace,
    )
    package_dir = os.path.join(output_dir, addon_name)
    shutil.copy(
        os.path.join(internal_source_dir, "static", "alerticon.png"),
        os.path.join(package_dir, "appserver", "static"),
    )
    schema_content = envs["schema.content"]

    conf_gen = alert_actions_conf_gen.AlertActionsConfGeneration(
        input_setting=schema_content,
        package_path=package_dir,
    )
    conf_gen.handle()

    html_gen = alert_actions_html_gen.AlertHtmlGenerator(
        input_setting=schema_content,
        package_path=package_dir,
    )
    html_gen.handle()

    py_gen = alert_actions_py_gen.AlertActionsPyGenerator(
        addon_name=addon_name,
        input_setting=schema_content,
        package_path=package_dir,
    )
    py_gen.handle()



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import re
from typing import Any, Dict, List
import warnings

import jsonschema

from splunk_add_on_ucc_framework import dashboard as dashboard_lib
from splunk_add_on_ucc_framework import global_config as global_config_lib


class GlobalConfigValidatorException(Exception):
    pass


class GlobalConfigValidator:
    """
    GlobalConfigValidator implements different validation for globalConfig file.
    Custom validation should be implemented here.
    """

    def __init__(self, source_dir: str, global_config: global_config_lib.GlobalConfig):
        self._source_dir = source_dir
        self._config = global_config.content

    def _validate_config_against_schema(self) -> None:
        """
        Validates config against JSON schema.
        Raises jsonschema.ValidationError if config is not valid.
        """
        schema_path = os.path.join(self._source_dir, "schema", "schema.json")
        with open(schema_path) as f_schema:
            schema_raw = f_schema.read()
            schema = json.loads(schema_raw)
        try:
            return jsonschema.validate(instance=self._config, schema=schema)
        except jsonschema.ValidationError as e:
            raise GlobalConfigValidatorException(e.message)

    def _validate_configuration_tab_table_has_name_field(self) -> None:
        """
        Validates that if a configuration tab should be rendered as a table,
        then it needs to have an entity which has field "name".
        """
        pages = self._config["pages"]
        configuration = pages["configuration"]
        tabs = configuration["tabs"]
        for tab in tabs:
            if "table" in tab:
                entities = tab["entity"]
                has_name_field = False
                for entity in entities:
                    if entity["field"] == "name":
                        has_name_field = True
                        break
                if not has_name_field:
                    raise GlobalConfigValidatorException(
                        f"Tab '{tab['name']}' should have entity with field 'name'"
                    )

    def _validate_custom_rest_handlers(self) -> None:
        """
        Validates that only "restHandlerName" or both "restHandlerModule" and
        "restHandlerClass" is present in the input configuration. Also validates
        that both "restHandlerModule" and "restHandlerClass" is present if any
        of them are present.

        The valid scenarios:
            * only restHandlerName is present
            * both restHandlerModule and restHandlerClass is present
        Everything other combination is considered invalid.
        """
        pages = self._config["pages"]
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            rest_handler_name = service.get("restHandlerName")
            rest_handler_module = service.get("restHandlerModule")
            rest_handler_class = service.get("restHandlerClass")
            if rest_handler_name is not None and (
                rest_handler_module is not None or rest_handler_class is not None
            ):
                raise GlobalConfigValidatorException(
                    f"Input '{service['name']}' has both 'restHandlerName' and "
                    f"'restHandlerModule' or 'restHandlerClass' fields present. "
                    f"Please use only 'restHandlerName' or 'restHandlerModule' "
                    f"and 'restHandlerClass'."
                )
            if (rest_handler_module is not None and rest_handler_class is None) or (
                rest_handler_module is None and rest_handler_class is not None
            ):
                raise GlobalConfigValidatorException(
                    f"Input '{service['name']}' should have both "
                    f"'restHandlerModule' and 'restHandlerClass' fields "
                    f"present, only 1 of them was found."
                )

    def _validate_file_type_entity(self) -> None:
        """
        Validates that file-based field has all necessary fields.
        Things should be provided in case of file field:
            * options field
            * supportedFileTypes field should be present in options
        """
        pages = self._config["pages"]
        configuration = pages["configuration"]
        tabs = configuration["tabs"]
        for tab in tabs:
            entities = tab["entity"]
            for entity in entities:
                if entity["type"] == "file":
                    options = entity.get("options")
                    if options is None:
                        raise GlobalConfigValidatorException(
                            f"Options field for the file type should be present "
                            f"for '{entity['field']}' field."
                        )
                    supported_file_types = options.get("supportedFileTypes")
                    if supported_file_types is None:
                        raise GlobalConfigValidatorException(
                            f"You should define your supported file types in "
                            f"the `supportedFileTypes` field for the "
                            f"'{entity['field']}' field."
                        )

    def _validate_string_validator(
        self, entity_field: str, validator: Dict[str, Any]
    ) -> None:
        """
        Validates string validator. maxLength should be greater or equal
        than minLength.
        """
        if validator["maxLength"] < validator["minLength"]:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect string validator, "
                f"'maxLength' should be greater or equal than 'minLength'."
            )

    def _validate_number_validator(
        self, entity_field: str, validator: Dict[str, Any]
    ) -> None:
        """
        Validates number validator, both values in range should be numbers and
        first one should be smaller than the second one.
        """
        validator_range = validator["range"]
        if len(validator_range) != 2:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect number validator, "
                f"it should have 2 elements under 'range' field."
            )
        # Schema already checks that the values in the 'range' field are
        # numbers.
        if validator_range[1] < validator_range[0]:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect number validator, "
                f"second element should be greater or equal than first element."
            )

    def _validate_regex_validator(
        self, entity_field: str, validator: Dict[str, Any]
    ) -> None:
        """
        Validates regex validator, provided regex should at least be compilable.
        """
        try:
            re.compile(validator["pattern"])
        except re.error:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect regex validator, "
                f"pattern provided in the 'pattern' field is not compilable."
            )

    def _validate_entity_validators(self, entity: Dict[str, Any]) -> None:
        """
        Validates entity validators.
        """
        validators = entity.get("validators", [])
        for validator in validators:
            if validator["type"] == "string":
                self._validate_string_validator(entity["field"], validator)
            if validator["type"] == "number":
                self._validate_number_validator(entity["field"], validator)
            if validator["type"] == "regex":
                self._validate_regex_validator(entity["field"], validator)

    def _validate_validators(self) -> None:
        """
        Validates both configuration and services validators, currently string,
        number and regex are supported.
        """
        pages = self._config["pages"]
        configuration = pages["configuration"]
        tabs = configuration["tabs"]
        for tab in tabs:
            entities = tab["entity"]
            for entity in entities:
                self._validate_entity_validators(entity)

        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            entities = service["entity"]
            for entity in entities:
                self._validate_entity_validators(entity)

    @staticmethod
    def _find_duplicates_in_list(_list: List[Any]) -> bool:
        return len(set(_list)) != len(_list)

    def _validate_children_duplicates(
        self, children: List[Dict[str, Any]], entity_label: str
    ) -> None:
        """
        Validates duplicates under children key in autoCompleteFields
        for fields under keys: label, value
        """
        labels, values = [], []
        for child in children:
            labels.append(child["label"].lower())
            values.append(child["value"].lower())
        if self._find_duplicates_in_list(values) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                f"Duplicates found for autoCompleteFields children in entity '{entity_label}'"
            )

    def _validate_autoCompleteFields_duplicates(
        self, options: Dict[str, Any], entity_label: str
    ) -> None:
        """
        Validates duplicates in autoCompleteFields keys
        for fields under keys: label, value
        If autoCompleteFields has children key, children validator is called
        """
        labels, values = [], []
        for field in options["autoCompleteFields"]:
            labels.append(field.get("label").lower())
            children = field.get("children")
            if children:
                self._validate_children_duplicates(children, entity_label)
            else:
                values.append(field.get("value").lower())
        if self._find_duplicates_in_list(values) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                f"Duplicates found for autoCompleteFields: '{entity_label}'"
            )

    def _validate_multilevel_menu(self) -> None:
        inputs = self._config.get("pages").get("inputs")
        if inputs:
            groups_menu = inputs.get("groupsMenu")
            if groups_menu:
                names, titles = [], []
                for group in groups_menu:
                    names.append(group.get("groupName").lower())
                    titles.append(group.get("groupTitle").lower())
                    group_services = group.get("groupServices")
                    if group_services:
                        for serviceName in group_services:
                            if not any(
                                serviceName == service.get("name")
                                for service in inputs.get("services")
                            ):
                                raise GlobalConfigValidatorException(
                                    f"{serviceName} ServiceName in the "
                                    f"multi-level menu does not match any "
                                    f"services name."
                                )
                    else:
                        if not any(
                            group.get("groupName") == service.get("name")
                            and group.get("groupTitle") == service.get("title")
                            for service in inputs.get("services")
                        ):
                            raise GlobalConfigValidatorException(
                                f'{group.get("groupName")} groupName or '
                                f'{group.get("groupTitle")} groupTitle in the '
                                f"multi-level menu does not match any services "
                                f"name or title."
                            )

                if self._find_duplicates_in_list(
                    names
                ) or self._find_duplicates_in_list(titles):
                    raise GlobalConfigValidatorException(
                        "Duplicates found for multi-level menu groups' names or titles."
                    )

    def _validate_entity_duplicates(self, entity: List[Dict[str, Any]]) -> None:
        """
        Validates duplicates in entity keys
        for fields under keys: field, label
        If entity has autoCompleteFields key, autoCompleteFields validator is called
        """
        fields, labels = [], []
        for _entity in entity:
            fields.append(_entity["field"].lower())
            labels.append(_entity["label"].lower())
            options = _entity.get("options")
            if options and options.get("autoCompleteFields"):
                self._validate_autoCompleteFields_duplicates(
                    _entity["options"], _entity["label"]
                )
        if self._find_duplicates_in_list(fields) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for entity field or label"
            )

    def _validate_tabs_duplicates(self, tabs: List[Dict[str, Any]]) -> None:
        """
        Validates duplicates in tab keys under configuration
        for fields under keys: name, title
        Calls for entity validator, as at least one entity is required in schema
        """
        names, titles = [], []
        for tab in tabs:
            names.append(tab["name"].lower())
            titles.append(tab["title"].lower())

            self._validate_entity_duplicates(tab["entity"])
        if self._find_duplicates_in_list(names) or self._find_duplicates_in_list(
            titles
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for tabs names or titles"
            )

    def _validate_inputs_duplicates(self, inputs: Dict[str, Any]) -> None:
        """
        Validates duplicates in tab keys under configuration
        for fields under keys: name, title
        Inputs are not required in schema
        """
        names, titles = [], []
        for service in inputs["services"]:
            names.append(service["name"].lower())
            titles.append(service["title"].lower())

            self._validate_entity_duplicates(service["entity"])

        if self._find_duplicates_in_list(names) or self._find_duplicates_in_list(
            titles
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for inputs (services) names or titles"
            )

    def _validate_duplicates(self) -> None:
        """
        Validates duplicates for both tabs and services (inputs). Inputs however are
        not required in schema, so this checks if globalConfig has inputs
        """
        pages = self._config["pages"]

        self._validate_tabs_duplicates(pages["configuration"]["tabs"])

        inputs = pages.get("inputs")
        if inputs:
            self._validate_inputs_duplicates(inputs)

    def _validate_alerts(self) -> None:
        # TODO: test cases should be added here.
        alerts = self._config.get("alerts", [])
        for alert in alerts:
            fields = []
            alert_entity = alert.get("entity")
            if alert_entity is None:
                continue
            for entity in alert_entity:
                if entity.get("field") in fields:
                    raise GlobalConfigValidatorException(
                        "Field names should be unique across alerts"
                    )
                else:
                    fields.append(entity.get("field"))
                entity_type = entity.get("type")
                if entity_type in ("radio", "singleSelect"):
                    if not entity.get("options"):
                        raise GlobalConfigValidatorException(
                            f"{entity_type} type must have options parameter"
                        )
                elif (
                    entity.get("options") and entity_type != "singleSelectSplunkSearch"
                ):
                    raise GlobalConfigValidatorException(
                        f"{entity_type} type must not contain options parameter"
                    )
                if entity_type in ("singleSelectSplunkSearch",):
                    if not all(
                        [
                            entity.get("search"),
                            entity.get("valueField"),
                            entity.get("labelField"),
                        ]
                    ):
                        raise GlobalConfigValidatorException(
                            f"{entity_type} type must have search, valueLabel and valueField parameters"
                        )
                elif any(
                    [
                        entity.get("search"),
                        entity.get("valueField"),
                        entity.get("labelField"),
                    ]
                ):
                    raise GlobalConfigValidatorException(
                        f"{entity_type} type must not contain search, valueField or labelField parameter"
                    )

    def _validate_panels(self) -> None:
        """
        Validates if the panels defined in the configuration are supported.
        """
        dashboard = self._config["pages"].get("dashboard")
        if dashboard:
            for panel in dashboard["panels"]:
                if panel["name"] not in dashboard_lib.SUPPORTED_PANEL_NAMES:
                    raise GlobalConfigValidatorException(
                        f"'{panel['name']}' is not a supported panel name. "
                        f"Supported panel names: {dashboard_lib.SUPPORTED_PANEL_NAMES_READABLE}"
                    )

    def _warn_on_placeholder_usage(self) -> None:
        """
        Warns if placeholder is used.
        More details here: https://github.com/splunk/addonfactory-ucc-generator/issues/831.
        """
        pages = self._config["pages"]
        configuration = pages["configuration"]
        tabs = configuration["tabs"]
        for tab in tabs:
            for entity in tab["entity"]:
                if "placeholder" in entity.get("options", {}):
                    warnings.warn(
                        f"`placeholder` option found for configuration tab '{tab['name']}' "
                        f"-> entity field '{entity['field']}'. "
                        f"Please take a look at https://github.com/splunk/addonfactory-ucc-generator/issues/831.",
                        DeprecationWarning,
                    )
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            for entity in service["entity"]:
                if "placeholder" in entity.get("options", {}):
                    warnings.warn(
                        f"`placeholder` option found for input service '{service['name']}' "
                        f"-> entity field '{entity['field']}'. "
                        f"Please take a look at https://github.com/splunk/addonfactory-ucc-generator/issues/831.",
                        DeprecationWarning,
                    )

    def _validate_checkbox_group(self) -> None:
        pages = self._config["pages"]
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            for entity in service["entity"]:
                if entity["type"] == "checkboxGroup":
                    row_field_names = []
                    for row in entity["options"]["rows"]:
                        if row["field"] in row_field_names:
                            raise GlobalConfigValidatorException(
                                f"Entity {entity['field']} has duplicate field ({row['field']}) in options.rows"
                            )
                        row_field_names.append(row["field"])
                    groups = entity["options"].get("groups")
                    if groups is None:
                        return
                    group_used_field_names = []
                    for group in groups:
                        for group_field_name in group["fields"]:
                            if group_field_name not in row_field_names:
                                raise GlobalConfigValidatorException(
                                    f"Entity {entity['field']} uses field ({group_field_name}) "
                                    f"which is not defined in options.rows"
                                )
                            if group_field_name in group_used_field_names:
                                raise GlobalConfigValidatorException(
                                    f"Entity {entity['field']} has duplicate field ({group_field_name}) "
                                    f"in options.groups"
                                )
                            group_used_field_names.append(group_field_name)

    def validate(self) -> None:
        self._validate_config_against_schema()
        self._validate_configuration_tab_table_has_name_field()
        self._validate_custom_rest_handlers()
        self._validate_file_type_entity()
        self._validate_validators()
        self._validate_multilevel_menu()
        self._validate_duplicates()
        self._validate_alerts()
        self._validate_panels()
        self._warn_on_placeholder_usage()
        self._validate_checkbox_group()



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
import logging
import os
import sys
import re

import shutil

from splunk_add_on_ucc_framework import utils

logger = logging.getLogger("ucc_gen")


ADDON_NAME_RE_STR = r'^[^<>:"/|?*]+$'
ADDON_NAME_RE = re.compile(ADDON_NAME_RE_STR)
ADDON_REST_ROOT_RE_STR = r"^\w+$"
ADDON_REST_ROOT_RE = re.compile(ADDON_REST_ROOT_RE_STR)
ADDON_INPUT_NAME_RE_STR = r"^[0-9a-zA-Z][\w-]*$"
ADDON_INPUT_NAME_RE = re.compile(ADDON_INPUT_NAME_RE_STR)


def _is_valid(pattern: re.Pattern[str], string: str) -> bool:
    result = pattern.search(string)
    if result is None:
        return False
    return True


def _is_valid_addon_name(addon_name: str) -> bool:
    return _is_valid(ADDON_NAME_RE, addon_name)


def _is_valid_rest_root_name(rest_root: str) -> bool:
    return _is_valid(ADDON_REST_ROOT_RE, rest_root)


def _is_valid_input_name(input_name: str) -> bool:
    if len(input_name) > 50:
        return False
    return _is_valid(ADDON_INPUT_NAME_RE, input_name)


def _generate_addon(
    addon_name: str,
    addon_display_name: str,
    addon_input_name: str,
    addon_version: str,
    addon_rest_root: str | None = None,
    overwrite: bool = False,
) -> str:
    generated_addon_path = os.path.join(
        os.getcwd(),
        addon_name,
    )
    if not overwrite and os.path.exists(generated_addon_path):
        logger.error(
            "The location is already taken, use `--overwrite` "
            "option to overwrite the content of existing folder."
        )
        sys.exit(1)
    if overwrite:
        try:
            shutil.rmtree(generated_addon_path)
        except FileNotFoundError:
            pass
    os.makedirs(generated_addon_path)
    readme_path = os.path.join(generated_addon_path, "README.md")
    with open(readme_path, "w") as _f:
        _f.write(f"# {addon_name}\n")
    global_config_path = os.path.join(generated_addon_path, "globalConfig.json")
    global_config_rendered_content = (
        utils.get_j2_env()
        .get_template("globalConfig.json.init-template")
        .render(
            addon_name=addon_name,
            addon_rest_root=addon_rest_root,
            addon_version=addon_version,
            addon_display_name=addon_display_name,
            addon_input_name=addon_input_name,
        )
    )
    with open(global_config_path, "w") as _f:
        _f.write(global_config_rendered_content)
    package_path = os.path.join(generated_addon_path, "package")
    os.makedirs(package_path)
    package_license_path = os.path.join(package_path, "LICENSE.txt")
    with open(package_license_path, "w") as _f:
        pass
    package_readme_path = os.path.join(package_path, "README.txt")
    with open(package_readme_path, "w") as _f:
        pass
    package_app_manifest_path = os.path.join(package_path, "app.manifest")
    app_manifest_rendered_content = (
        utils.get_j2_env()
        .get_template("app.manifest.init-template")
        .render(
            addon_name=addon_name,
            addon_version=addon_version,
            addon_display_name=addon_display_name,
        )
    )
    with open(package_app_manifest_path, "w") as _f:
        _f.write(app_manifest_rendered_content)
    package_bin_path = os.path.join(package_path, "bin")
    os.makedirs(package_bin_path)
    package_bin_input_path = os.path.join(package_bin_path, f"{addon_input_name}.py")
    input_rendered_content = (
        utils.get_j2_env()
        .get_template("input.name.init-template")
        .render(
            addon_name=addon_name,
            addon_rest_root=addon_rest_root,
            addon_input_name=addon_input_name,
            addon_version=addon_version,
            addon_display_name=addon_display_name,
        )
    )
    with open(package_bin_input_path, "w") as _f:
        _f.write(input_rendered_content + "\n")
    package_lib_path = os.path.join(package_path, "lib")
    os.makedirs(package_lib_path)
    package_lib_requirements_path = os.path.join(package_lib_path, "requirements.txt")
    with open(package_lib_requirements_path, "w") as _f:
        _f.writelines(
            [
                "splunktaucclib\n",
                "splunk-sdk\n",
                "solnlib\n",
            ]
        )
    return generated_addon_path


def init(
    addon_name: str,
    addon_display_name: str,
    addon_input_name: str,
    addon_version: str,
    addon_rest_root: str | None = None,
    overwrite: bool = False,
) -> str:
    if not _is_valid_addon_name(addon_name):
        logger.error(
            f"Add-on name provided is not valid, it should follow '{ADDON_NAME_RE_STR}' regex."
        )
        sys.exit(1)

    if addon_rest_root is None:
        if _is_valid_rest_root_name(addon_name):
            logger.info(
                "Parameter `--addon-rest-root` is not provided, using `--addon-name` as a REST root."
            )
            addon_rest_root = addon_name
        else:
            logger.error(
                f"Can not use add-on name provided as an add-on REST root. "
                f"It should follow '{ADDON_REST_ROOT_RE_STR}' regex. "
                f"Please provide `--addon-rest-root` to specify add-on REST root."
            )
            sys.exit(1)
    else:
        if not _is_valid_rest_root_name(addon_rest_root):
            logger.error(
                f"Add-on REST root provided is not valid, it should follow '{ADDON_REST_ROOT_RE_STR}' regex."
            )
            sys.exit(1)
    if not _is_valid_input_name(addon_input_name):
        logger.error(
            f"Add-on input name provided is not valid, "
            f"it should follow '{ADDON_INPUT_NAME_RE_STR}' regex and be less than 50 characters."
        )
        sys.exit(1)
    generated_addon_path = _generate_addon(
        addon_name,
        addon_display_name,
        addon_input_name,
        addon_version,
        addon_rest_root,
        overwrite,
    )
    logger.info(f"Generated add-on is located here {generated_addon_path}")
    logger.info(
        "LICENSE.txt and README.txt are empty, "
        "you may need to modify the content of those files. "
    )
    return generated_addon_path



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import Any, Dict, Tuple, List

from splunk_add_on_ucc_framework import global_config as global_config_lib

logger = logging.getLogger("ucc_gen")


def _version_tuple(version: str) -> Tuple[str, ...]:
    """
    convert string into tuple to compare version

    Args:
        version_str : raw string
    Returns:
        tuple : version into tupleformat
    """
    filled = []
    for point in version.split("."):
        filled.append(point.zfill(8))
    return tuple(filled)


def _handle_biased_terms(conf_entities: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    for entity in conf_entities:
        entity_option = entity.get("options")
        if entity_option and "whiteList" in entity_option:
            entity_option["allowList"] = entity_option.get("whiteList")
            del entity_option["whiteList"]
        if entity_option and "blackList" in entity_option:
            entity_option["denyList"] = entity_option.get("blackList")
            del entity_option["blackList"]
    return conf_entities


def _handle_biased_terms_update(global_config: global_config_lib.GlobalConfig) -> None:
    for tab in global_config.tabs:
        conf_entities = tab.get("entity")
        tab["entity"] = _handle_biased_terms(conf_entities)
    for service in global_config.inputs:
        conf_entities = service.get("entity")
        service["entity"] = _handle_biased_terms(conf_entities)
    global_config.update_schema_version("0.0.1")


def _handle_dropping_api_version_update(
    global_config: global_config_lib.GlobalConfig,
) -> None:
    if global_config.meta.get("apiVersion"):
        del global_config.meta["apiVersion"]
    global_config.update_schema_version("0.0.3")


def handle_global_config_update(global_config: global_config_lib.GlobalConfig) -> None:
    """Handle changes in globalConfig file."""
    current_schema_version = global_config.schema_version
    version = current_schema_version if current_schema_version else "0.0.0"
    logger.info(f"Current globalConfig schema version is {current_schema_version}")

    if _version_tuple(version) < _version_tuple("0.0.1"):
        _handle_biased_terms_update(global_config)
        global_config.dump(global_config.original_path)
        logger.info("Updated globalConfig schema to version 0.0.1")

    if _version_tuple(version) < _version_tuple("0.0.2"):
        for tab in global_config.tabs:
            if tab["name"] == "account":
                conf_entities = tab.get("entity")
                oauth_state_enabled_entity = {}
                for entity in conf_entities:
                    if entity.get("field") == "oauth_state_enabled":
                        logger.warning(
                            "oauth_state_enabled field is no longer a separate "
                            "entity since UCC version 5.0.0. It is now an "
                            "option in the oauth field. Please update the "
                            "globalConfig file accordingly."
                        )
                        oauth_state_enabled_entity = entity

                    if entity.get("field") == "oauth" and not entity.get(
                        "options", {}
                    ).get("oauth_state_enabled"):
                        entity["options"]["oauth_state_enabled"] = False

                if oauth_state_enabled_entity:
                    conf_entities.remove(oauth_state_enabled_entity)

            tab_options = tab.get("options", {})
            if tab_options.get("onChange"):
                logger.error(
                    "The onChange option is no longer supported since UCC "
                    "version 5.0.0. You can use custom hooks to implement "
                    "these actions."
                )
                del tab_options["onChange"]
            if tab_options.get("onLoad"):
                logger.error(
                    "The onLoad option is no longer supported since UCC "
                    "version 5.0.0. You can use custom hooks to implement "
                    "these actions."
                )
                del tab_options["onLoad"]

        if global_config.has_inputs():
            for service in global_config.inputs:
                service_options = service.get("options", {})
                if service_options.get("onChange"):
                    logger.error(
                        "The onChange option is no longer supported since UCC "
                        "version 5.0.0. You can use custom hooks to implement "
                        "these actions."
                    )
                    del service_options["onChange"]
                if service_options.get("onLoad"):
                    logger.error(
                        "The onLoad option is no longer supported since UCC "
                        "version 5.0.0. You can use custom hooks to implement "
                        "these actions."
                    )
                    del service_options["onLoad"]
        global_config.update_schema_version("0.0.2")
        global_config.dump(global_config.original_path)
        logger.info("Updated globalConfig schema to version 0.0.2")

    if _version_tuple(version) < _version_tuple("0.0.3"):
        _handle_dropping_api_version_update(global_config)
        global_config.dump(global_config.original_path)
        logger.info("Updated globalConfig schema to version 0.0.3")



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
from typing import Dict, List, Set, Any

from splunk_add_on_ucc_framework import global_config as global_config_lib

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.datainput import (
    DataInputEndpointBuilder,
    DataInputEntityBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.field import (
    RestFieldBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.multiple_model import (
    MultipleModelEndpointBuilder,
    MultipleModelEntityBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.oauth_model import (
    OAuthModelEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.single_model import (
    SingleModelEndpointBuilder,
    SingleModelEntityBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.validator_builder import (
    ValidatorBuilder,
)

REST_HANDLER_DEFAULT_MODULE = "splunktaucclib.rest_handler.admin_external"
REST_HANDLER_DEFAULT_CLASS = "AdminExternalHandler"


def _is_true(val: Any) -> bool:
    return str(val).strip().upper() in ("1", "TRUE", "T", "Y", "YES")


class GlobalConfigBuilderSchema:
    def __init__(self, global_config: global_config_lib.GlobalConfig):
        self.global_config = global_config
        self._settings_conf_file_names: Set[str] = set()
        self._configs_conf_file_names: Set[str] = set()
        self._oauth_conf_file_names: Set[str] = set()
        self._endpoints: Dict[str, RestEndpointBuilder] = {}
        self._parse_builder_schema()

    @property
    def product(self) -> str:
        return self.global_config.product

    @property
    def namespace(self) -> str:
        return self.global_config.namespace

    @property
    def settings_conf_file_names(self) -> Set[str]:
        return self._settings_conf_file_names

    @property
    def configs_conf_file_names(self) -> Set[str]:
        return self._configs_conf_file_names

    @property
    def oauth_conf_file_names(self) -> Set[str]:
        return self._oauth_conf_file_names

    @property
    def endpoints(self) -> List[RestEndpointBuilder]:
        return list(self._endpoints.values())

    def _parse_builder_schema(self) -> None:
        self._builder_configs()
        self._builder_settings()
        self._builder_inputs()

    def _builder_configs(self) -> None:
        for config in self.global_config.configs:
            name = config["name"]
            endpoint = SingleModelEndpointBuilder(
                name=name,
                namespace=self.global_config.namespace,
                rest_handler_name=config.get("restHandlerName"),
                rest_handler_module=REST_HANDLER_DEFAULT_MODULE,
                rest_handler_class=REST_HANDLER_DEFAULT_CLASS,
            )
            self._endpoints[name] = endpoint
            content = self._get_oauth_enitities(config["entity"])
            fields = self._parse_fields(content)
            entity = SingleModelEntityBuilder(
                None,
                fields,
                conf_name=config.get("conf"),
            )
            endpoint.add_entity(entity)
            # If we have given oauth support then we have to add endpoint for accesstoken
            for entity_element in config["entity"]:
                if entity_element["type"] == "oauth":
                    oauth_endpoint = OAuthModelEndpointBuilder(
                        name="oauth",
                        namespace=self.global_config.namespace,
                        app_name=self.global_config.product,
                    )
                    self._endpoints["oauth"] = oauth_endpoint
                    self._oauth_conf_file_names.add(oauth_endpoint.conf_name)
            self._configs_conf_file_names.add(endpoint.conf_name)

    def _builder_settings(self) -> None:
        endpoint = MultipleModelEndpointBuilder(
            name="settings",
            namespace=self.global_config.namespace,
            rest_handler_module=REST_HANDLER_DEFAULT_MODULE,
            rest_handler_class=REST_HANDLER_DEFAULT_CLASS,
        )
        self._endpoints["settings"] = endpoint
        for setting in self.global_config.settings:
            content = self._get_oauth_enitities(setting["entity"])
            fields = self._parse_fields(content)
            entity = MultipleModelEntityBuilder(
                setting["name"],
                fields,
            )
            endpoint.add_entity(entity)
            self._settings_conf_file_names.add(endpoint.conf_name)

    def _builder_inputs(self) -> None:
        for input_item in self.global_config.inputs:
            rest_handler_name = input_item.get("restHandlerName")
            rest_handler_module = input_item.get(
                "restHandlerModule",
                REST_HANDLER_DEFAULT_MODULE,
            )
            rest_handler_class = input_item.get(
                "restHandlerClass",
                REST_HANDLER_DEFAULT_CLASS,
            )
            if "conf" in input_item:
                name = input_item["name"]
                single_model_endpoint = SingleModelEndpointBuilder(
                    name=name,
                    namespace=self.global_config.namespace,
                    rest_handler_name=rest_handler_name,
                    rest_handler_module=rest_handler_module,
                    rest_handler_class=rest_handler_class,
                )
                self._endpoints[name] = single_model_endpoint
                content = self._get_oauth_enitities(input_item["entity"])
                fields = self._parse_fields(content)
                single_model_entity = SingleModelEntityBuilder(
                    None,
                    fields,
                    conf_name=input_item["conf"],
                )
                single_model_endpoint.add_entity(single_model_entity)
            else:
                name = input_item["name"]
                data_input_endpoint = DataInputEndpointBuilder(
                    name=name,
                    namespace=self.global_config.namespace,
                    input_type=input_item["name"],
                    rest_handler_name=rest_handler_name,
                    rest_handler_module=rest_handler_module,
                    rest_handler_class=rest_handler_class,
                )
                self._endpoints[name] = data_input_endpoint
                content = self._get_oauth_enitities(input_item["entity"])
                fields = self._parse_fields(content)
                data_input_entity = DataInputEntityBuilder(
                    None,
                    fields,
                    input_type=input_item["name"],
                )
                data_input_endpoint.add_entity(data_input_entity)

    def _parse_fields(
        self, fields_content: List[Dict[str, Any]]
    ) -> List[RestFieldBuilder]:
        return [
            RestFieldBuilder(
                field["field"],
                _is_true(field.get("required")),
                _is_true(field.get("encrypted")),
                field.get("defaultValue"),
                ValidatorBuilder().build(field.get("validators")),
            )
            for field in fields_content
            if field["field"] != "name"
        ]

    """
    If the entity contains type oauth then we need to alter the content to generate proper entities to generate
    the rest handler with the oauth fields
    :param content: json content of entity
    :type content: `json`
    """

    def _get_oauth_enitities(
        self, content: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        for entity_element in content:
            # Check if we have oauth type
            if entity_element["type"] == "oauth":
                # Check if we have both basic and oauth type authentication is required
                if (
                    "basic" in entity_element["options"]["auth_type"]
                    and "oauth" in entity_element["options"]["auth_type"]
                ):
                    # Append all the basic auth fields to the content
                    content = content + entity_element["options"]["basic"]
                    # Append oauth auth fields to the content
                    content = content + entity_element["options"]["oauth"]
                    # Append auth_type, access_token, refresh_token & instance_url fields
                    content = content + json.loads(
                        '[{"field": "access_token","encrypted": true},'
                        '{"field": "refresh_token","encrypted":true},'
                        '{"field": "instance_url"},'
                        '{"field": "auth_type"}]'
                    )
                # If only oauth type authentication is required
                elif "oauth" in entity_element["options"]["auth_type"]:
                    # Append all the oauth auth fields to the content
                    content = content + entity_element["options"]["oauth"]
                    # Append access_token, refresh_token & instance_url fields
                    content = content + json.loads(
                        '[{"field": "access_token","encrypted": true},'
                        '{"field": "refresh_token","encrypted":true},'
                        '{"field": "instance_url"}]'
                    )
                # We will remove the oauth type entity as we have replaced it with all the entity fields
                content.remove(entity_element)
                break
        return content



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any, Dict, Optional, Sequence

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    indent,
)


def quote_regex(value: Optional[str]) -> str:
    return '"""%s"""' % value


class BaseValidator:
    _validation_template = """validator.{class_name}({arguments})"""

    def _get_class_name(self) -> str:
        raise NotImplementedError()

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError()

    def _format_arguments(self, **kwargs: Dict[str, Any]) -> str:
        args = list(
            map(
                lambda k_v: f"{k_v[0]}={k_v[1]}, ",
                list(kwargs.items()),
            )
        )
        args.insert(0, "")
        args.append("")
        return indent("\n".join(args))

    def build(self, config: Dict[str, Any]) -> str:
        return self._validation_template.format(
            class_name=self._get_class_name(),
            arguments=self._format_arguments(**self._get_arguments(config)),
        )


class StringValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "String"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        return {
            "max_len": config.get("maxLength"),
            "min_len": config.get("minLength"),
        }


class NumberValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Number"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        ranges = config.get("range", [None, None])
        return {
            "max_val": ranges[1],
            "min_val": ranges[0],
        }


class RegexValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Pattern"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        return {"regex": "r" + quote_regex(config.get("pattern"))}


class EmailValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Pattern"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        regex = (
            r"^[a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}"
            r"[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"
        )
        return {"regex": "r" + quote_regex(regex)}


class Ipv4Validator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Pattern"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        regex = r"^(?:(?:[0-1]?\d{1,2}|2[0-4]\d|25[0-5])(?:\.|$)){4}$"
        return {"regex": "r" + quote_regex(regex)}


class DateValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Pattern"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        # iso8601 date time format
        regex = (
            r"^\s*((?:[+-]\d{6}|\d{4})-(?:\d\d-\d\d|W\d\d-\d|W\d\d|\d\d\d|\d\d))"
            r"(?:(T| )(\d\d(?::\d\d(?::\d\d(?:[.,]\d+)?)?)?)([\+\-]\d\d(?::?\d\d)?|\s*Z)?)?$"
        )
        return {"regex": "r" + quote_regex(regex)}


class UrlValidator(BaseValidator):
    def _get_class_name(self) -> str:
        return "Pattern"

    def _get_arguments(self, config: Dict[str, Any]) -> Dict[str, Any]:
        regex = (
            r"^(?:(?:https?|ftp|opc\.tcp):\/\/)?(?:\S+(?::\S*)?@)?"
            r"(?:(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
            r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}"
            r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|"
            r"(?:(?:[a-z\u00a1-\uffff0-9]+-?_?)*[a-z\u00a1-\uffff0-9]+)"
            r"(?:\.(?:[a-z\u00a1-\uffff0-9]+-?)*[a-z\u00a1-\uffff0-9]+)*"
            r"(?:\.(?:[a-z\u00a1-\uffff]{2,}))?)(?::\d{2,5})?(?:\/[^\s]*)?$"
        )
        return {"regex": "r" + quote_regex(regex)}


class ValidatorBuilder:
    _validation_config_map = {
        "string": StringValidator,
        "number": NumberValidator,
        "regex": RegexValidator,
        "email": EmailValidator,
        "ipv4": Ipv4Validator,
        "date": DateValidator,
        "url": UrlValidator,
        # file validator does not need any generated code, everything is
        # validated in the UI
    }

    def _format_multiple_validators(self, validators: Sequence[str]) -> str:
        validators_str = ", \n".join(validators)
        return """validator.AllOf(\n{validators}\n)""".format(
            validators=indent(validators_str),
        )

    def build(self, configs: Optional[Sequence[Dict[str, Any]]]) -> Optional[str]:
        if configs is None:
            return None
        generated_validators = []
        for config in configs:
            # `config` variable should always have `type` field according to
            # the schema.
            config_type: str = config.get("type", "")
            validator = self._validation_config_map.get(config_type)
            if validator is None:
                continue
            generated_validators.append(validator().build(config))
        if not generated_validators:
            return None
        if len(generated_validators) > 1:
            return self._format_multiple_validators(generated_validators)
        return generated_validators[0]



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os.path as op
from os import makedirs, remove

import addonfactory_splunk_conf_parser_lib as conf_parser

from splunk_add_on_ucc_framework.commands.modular_alert_builder.alert_actions_merge import (
    merge_conf_file,
)

logger = logging.getLogger("ucc_gen")


def write_file(file_name: str, file_path: str, content: str) -> None:
    logger.debug('operation="write", object="%s" object_type="file"', file_path)

    do_merge = False
    if file_name.endswith(".conf") or file_name.endswith("conf.spec"):
        do_merge = True
    else:
        logger.debug(
            f"'{file_name}' is not going to be merged, only .conf and "
            f".conf.spec files are supported."
        )

    new_file = None
    if op.exists(file_path) and do_merge:
        new_file = op.join(op.dirname(file_path), "new_" + file_name)
    if new_file:
        try:
            with open(new_file, "w+") as fhandler:
                fhandler.write(content)
            merge_conf_file(new_file, file_path)
        finally:
            if op.exists(new_file):
                remove(new_file)
    else:
        if not op.exists(op.dirname(file_path)):
            makedirs(op.dirname(file_path))
        with open(file_path, "w+") as fhandler:
            fhandler.write(content)
        if do_merge:
            parser = conf_parser.TABConfigParser()
            parser.read(file_path)
            with open(file_path, "w") as df:
                parser.write(df)



import import_declare_test

import os
import os.path as op
import sys
import time
import datetime
import json

import traceback
import requests
from splunklib import modularinput as smi
from solnlib import conf_manager
from solnlib import log
from solnlib.modular_input import checkpointer
from splunktaucclib.modinput_wrapper import base_modinput as base_mi



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Sequence

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
)


class RestmapConf:
    _admin_template = """
[admin:{namespace}]
match = /
members = {endpoints}
"""

    _external_template = """
[admin_external:{name}]
handlertype = python
python.version = python3
handlerfile = {rh_name}.py
handleractions = {actions}
handlerpersistentmode = true
"""

    @classmethod
    def build(cls, endpoints: Sequence[RestEndpointBuilder], namespace: str) -> str:
        externals = [
            cls._admin_template.format(
                namespace=namespace,
                endpoints=", ".join([ep.name for ep in endpoints]),
            )
        ]
        for endpoint in endpoints:
            external = cls._external_template.format(
                name=endpoint.name,
                rh_name=endpoint.rh_name,
                actions=", ".join(endpoint.actions()),
            )
            externals.append(external)
        return "".join(externals)



#
# Copyright 2023 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import configparser
import json
import logging
import os
import shutil
import sys
from typing import Optional, List
import subprocess

from openapi3 import OpenAPI

from splunk_add_on_ucc_framework import (
    __version__,
    exceptions,
    global_config_update,
    global_config_validator,
    utils,
)
from splunk_add_on_ucc_framework import dashboard
from splunk_add_on_ucc_framework import app_conf as app_conf_lib
from splunk_add_on_ucc_framework import meta_conf as meta_conf_lib
from splunk_add_on_ucc_framework import server_conf as server_conf_lib
from splunk_add_on_ucc_framework import app_manifest as app_manifest_lib
from splunk_add_on_ucc_framework import global_config as global_config_lib
from splunk_add_on_ucc_framework import data_ui_generator
from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    builder as alert_builder,
)
from splunk_add_on_ucc_framework.commands.rest_builder import (
    global_config_builder_schema,
)
from splunk_add_on_ucc_framework.commands.rest_builder.builder import RestBuilder
from splunk_add_on_ucc_framework.install_python_libraries import (
    SplunktaucclibNotFound,
    install_python_libraries,
)
from splunk_add_on_ucc_framework.commands.openapi_generator import (
    ucc_to_oas,
)


logger = logging.getLogger("ucc_gen")

internal_root_dir = os.path.dirname(os.path.dirname(__file__))


def _modify_and_replace_token_for_oauth_templates(
    ta_name: str, global_config: global_config_lib.GlobalConfig, outputdir: str
) -> None:
    """
    Rename templates with respect to addon name if OAuth is configured.

    Args:
        ta_name: Add-on name.
        global_config: Object representing globalConfig.
        outputdir: output directory.
    """
    redirect_js_src = os.path.join(
        outputdir, ta_name, "appserver", "static", "js", "build", "redirect_page.js"
    )
    redirect_html_src = os.path.join(
        outputdir, ta_name, "appserver", "templates", "redirect.html"
    )

    if global_config.has_oauth():
        html_template_path = os.path.join(outputdir, ta_name, "appserver", "templates")
        with open(os.path.join(html_template_path, "redirect.html")) as f:
            s = f.read()

        with open(os.path.join(html_template_path, "redirect.html"), "w") as f:
            s = s.replace("${ta.name}", ta_name.lower())
            s = s.replace("${ta.version}", global_config.version)
            f.write(s)

        redirect_js_dest = (
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "")
            + ta_name.lower()
            + "_redirect_page."
            + global_config.version
            + ".js"
        )
        redirect_html_dest = os.path.join(
            outputdir,
            ta_name,
            "appserver",
            "templates",
            ta_name.lower() + "_redirect.html",
        )
        with open(
            os.path.join(
                outputdir,
                ta_name,
                "default",
                "data",
                "ui",
                "views",
                f"{ta_name.lower()}_redirect.xml",
            ),
            "w",
        ) as addon_redirect_xml_file:
            addon_redirect_xml_content = data_ui_generator.generate_views_redirect_xml(
                ta_name,
            )
            addon_redirect_xml_file.write(addon_redirect_xml_content)
        os.rename(redirect_js_src, redirect_js_dest)
        os.rename(redirect_html_src, redirect_html_dest)
    else:
        os.remove(redirect_html_src)
        os.remove(redirect_js_src)


def _add_modular_input(
    ta_name: str, global_config: global_config_lib.GlobalConfig, outputdir: str
) -> None:
    for service in global_config.inputs:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_allow_list = frozenset(["name", "interval", "index", "sourcetype"])
        template = "input.template"
        if "template" in service:
            template = service.get("template") + ".template"

        # filter fields in allow list
        entity = [x for x in entity if x.get("field") not in field_allow_list]

        content = (
            utils.get_j2_env()
            .get_template(template)
            .render(
                input_name=input_name,
                class_name=class_name,
                description=description,
                entity=entity,
            )
        )
        input_file_name = os.path.join(outputdir, ta_name, "bin", input_name + ".py")
        with open(input_file_name, "w") as input_file:
            input_file.write(content)

        input_default = os.path.join(outputdir, ta_name, "default", "inputs.conf")
        config = configparser.ConfigParser()
        if os.path.exists(input_default):
            config.read(input_default)

        if config.has_section(input_name):
            config[input_name]["python.version"] = "python3"
        else:
            config[input_name] = {"python.version": "python3"}

        with open(input_default, "w") as configfile:
            config.write(configfile)


def _get_ignore_list(
    addon_name: str, ucc_ignore_path: str, output_directory: str
) -> List[str]:
    """
    Return path of files/folders to be removed.

    Args:
        addon_name: Add-on name.
        ucc_ignore_path: Path to '.uccignore'.
        output_directory: Output directory path.

    Returns:
        list: List of paths to be removed from output directory.
    """
    if not os.path.exists(ucc_ignore_path):
        return []
    else:
        with open(ucc_ignore_path) as ignore_file:
            ignore_list = ignore_file.readlines()
        ignore_list = [
            (
                os.path.join(output_directory, addon_name, utils.get_os_path(path))
            ).strip()
            for path in ignore_list
        ]
        return ignore_list


def _remove_listed_files(ignore_list: List[str]) -> None:
    """
    Return path of files/folders to removed in output folder.

    Args:
        ignore_list (list): List of files/folder to removed in output directory.

    """
    for path in ignore_list:
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
        else:
            logger.info(
                "While ignoring the files mentioned in .uccignore {} was not found".format(
                    path
                )
            )


def generate_data_ui(
    output_directory: str,
    addon_name: str,
    include_inputs: bool,
    include_dashboard: bool,
) -> None:
    # Create directories in the output folder for add-on's UI nav and views.
    os.makedirs(
        os.path.join(output_directory, addon_name, "default", "data", "ui", "nav"),
        exist_ok=True,
    )
    os.makedirs(
        os.path.join(output_directory, addon_name, "default", "data", "ui", "views"),
        exist_ok=True,
    )
    default_ui_path = os.path.join(
        output_directory, addon_name, "default", "data", "ui"
    )
    default_xml_path = os.path.join(default_ui_path, "nav", "default.xml")
    if os.path.exists(default_xml_path):
        logger.info(
            "Skipping generating data/ui/nav/default.xml because file already exists."
        )
    else:
        with open(default_xml_path, "w") as default_xml_file:
            default_xml_content = data_ui_generator.generate_nav_default_xml(
                include_inputs=include_inputs,
                include_dashboard=include_dashboard,
            )
            default_xml_file.write(default_xml_content)
    with open(
        os.path.join(default_ui_path, "views", "configuration.xml"), "w"
    ) as configuration_xml_file:
        configuration_xml_content = data_ui_generator.generate_views_configuration_xml(
            addon_name,
        )
        configuration_xml_file.write(configuration_xml_content)
    if include_inputs:
        with open(
            os.path.join(default_ui_path, "views", "inputs.xml"), "w"
        ) as input_xml_file:
            inputs_xml_content = data_ui_generator.generate_views_inputs_xml(
                addon_name,
            )
            input_xml_file.write(inputs_xml_content)


def _get_addon_version(addon_version: Optional[str]) -> str:
    if not addon_version:
        try:
            return utils.get_version_from_git()
        except exceptions.IsNotAGitRepo:
            logger.error(
                "Could not determine a version using Git (maybe not a Git "
                "repository?). Use `--ta-version` to specify the version you "
                "want."
            )
            exit(1)
        except exceptions.CouldNotVersionFromGitException:
            logger.error(
                "Could not find the proper version from git tags. "
                "Check out "
                "https://github.com/splunk/addonfactory-ucc-generator/issues/404"
            )
            exit(1)
    return addon_version.strip()


def _get_app_manifest(source: str) -> app_manifest_lib.AppManifest:
    app_manifest_path = os.path.abspath(
        os.path.join(source, app_manifest_lib.APP_MANIFEST_FILE_NAME),
    )
    with open(app_manifest_path) as manifest_file:
        app_manifest_content = manifest_file.read()
    try:
        app_manifest = app_manifest_lib.AppManifest(app_manifest_content)
        app_manifest.validate()
        return app_manifest
    except app_manifest_lib.AppManifestFormatException as e:
        logger.error(
            f"Manifest file @ {app_manifest_path} has invalid format.\n"
            f"Please refer to {app_manifest_lib.APP_MANIFEST_WEBSITE}.\n"
            f"Error message: {e}.\n"
        )
        sys.exit(1)


def _get_build_output_path(output_directory: Optional[str] = None) -> str:
    if output_directory is None:
        return os.path.join(os.getcwd(), "output")
    else:
        if not os.path.isabs(output_directory):
            return os.path.join(os.getcwd(), output_directory)
        return output_directory


def _get_python_version_from_executable(python_binary_name: str) -> str:
    try:
        python_binary_version = subprocess.run(
            [python_binary_name, "--version"], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")

        return python_binary_version.strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        raise exceptions.CouldNotIdentifyPythonVersionException(
            f"Failed to identify python version for binary {python_binary_name}"
        )


def generate(
    source: str,
    config_path: Optional[str] = None,
    addon_version: Optional[str] = None,
    output_directory: Optional[str] = None,
    python_binary_name: str = "python3",
) -> None:
    logger.info(f"ucc-gen version {__version__} is used")
    logger.info(f"Python binary name to use: {python_binary_name}")

    try:
        python_binary_version = _get_python_version_from_executable(python_binary_name)
        logger.info(f"Python Version: {python_binary_version}")
    except exceptions.CouldNotIdentifyPythonVersionException as e:
        logger.error(
            f"Failed to identify Python version for library installation. Error: {e}"
        )
        sys.exit(1)

    output_directory = _get_build_output_path(output_directory)
    logger.info(f"Output folder is {output_directory}")
    addon_version = _get_addon_version(addon_version)
    logger.info(f"Add-on will be built with version '{addon_version}'")
    if not os.path.exists(source):
        raise NotADirectoryError(f"{os.path.abspath(source)} not found.")
    shutil.rmtree(os.path.join(output_directory), ignore_errors=True)
    os.makedirs(os.path.join(output_directory))
    logger.info(f"Cleaned out directory {output_directory}")
    app_manifest = _get_app_manifest(source)
    ta_name = app_manifest.get_addon_name()
    if not config_path:
        is_global_config_yaml = False
        config_path = os.path.abspath(
            os.path.join(source, os.pardir, "globalConfig.json")
        )
        if not os.path.isfile(config_path):
            config_path = os.path.abspath(
                os.path.join(source, os.pardir, "globalConfig.yaml")
            )
            is_global_config_yaml = True
    else:
        is_global_config_yaml = True if config_path.endswith(".yaml") else False

    if os.path.isfile(config_path):
        logger.info(f"Using globalConfig file located @ {config_path}")
        global_config = global_config_lib.GlobalConfig(
            config_path, is_global_config_yaml
        )
        try:
            validator = global_config_validator.GlobalConfigValidator(
                internal_root_dir, global_config
            )
            validator.validate()
            logger.info("globalConfig file is valid")
        except global_config_validator.GlobalConfigValidatorException as e:
            logger.error(f"globalConfig file is not valid. Error: {e}")
            sys.exit(1)
        global_config.update_addon_version(addon_version)
        global_config.dump(global_config.original_path)
        logger.info(
            f"Updated and saved add-on version in the globalConfig file to {addon_version}"
        )
        global_config_update.handle_global_config_update(global_config)
        scheme = global_config_builder_schema.GlobalConfigBuilderSchema(global_config)
        utils.recursive_overwrite(
            os.path.join(internal_root_dir, "package"),
            os.path.join(output_directory, ta_name),
        )
        generate_data_ui(
            output_directory,
            ta_name,
            global_config.has_inputs(),
            global_config.has_dashboard(),
        )
        logger.info("Copied UCC template directory")
        global_config_file = (
            "globalConfig.yaml" if is_global_config_yaml else "globalConfig.json"
        )
        output_global_config_path = os.path.join(
            output_directory,
            ta_name,
            "appserver",
            "static",
            "js",
            "build",
            global_config_file,
        )
        shutil.copyfile(
            config_path,
            output_global_config_path,
        )
        logger.info("Copied globalConfig to output")
        ucc_lib_target = os.path.join(output_directory, ta_name, "lib")
        try:
            install_python_libraries(
                source, ucc_lib_target, python_binary_name, includes_ui=True
            )
        except SplunktaucclibNotFound as e:
            logger.error(str(e))
            sys.exit(1)
        logger.info(
            f"Installed add-on requirements into {ucc_lib_target} from {source}"
        )
        builder_obj = RestBuilder(scheme, os.path.join(output_directory, ta_name))
        builder_obj.build()
        _modify_and_replace_token_for_oauth_templates(
            ta_name,
            global_config,
            output_directory,
        )
        if global_config.has_inputs():
            logger.info("Generating inputs code")
            _add_modular_input(ta_name, global_config, output_directory)
        if global_config.has_alerts():
            logger.info("Generating alerts code")
            alert_builder.generate_alerts(
                global_config, ta_name, internal_root_dir, output_directory
            )

        conf_file_names = []
        conf_file_names.extend(list(scheme.settings_conf_file_names))
        conf_file_names.extend(list(scheme.configs_conf_file_names))
        conf_file_names.extend(list(scheme.oauth_conf_file_names))

        source_server_conf_path = os.path.join(source, "default", "server.conf")
        # For now, only create server.conf only if no server.conf is present in
        # the source package.
        if not os.path.isfile(source_server_conf_path):
            server_conf = server_conf_lib.ServerConf()
            server_conf.create_default(conf_file_names)
            output_server_conf_path = os.path.join(
                output_directory,
                ta_name,
                "default",
                server_conf_lib.SERVER_CONF_FILE_NAME,
            )
            server_conf.write(output_server_conf_path)
            logger.info(
                f"Created default {server_conf_lib.SERVER_CONF_FILE_NAME} file in the output folder"
            )
        if global_config.has_dashboard():
            logger.info("Including dashboard")
            dashboard_xml_path = os.path.join(
                output_directory,
                ta_name,
                "default",
                "data",
                "ui",
                "views",
                "dashboard.xml",
            )
            dashboard.generate_dashboard(
                global_config,
                ta_name,
                dashboard_xml_path,
            )
    else:
        global_config = None
        conf_file_names = []
        logger.warning(
            "Skipped generating UI components as globalConfig file does not exist"
        )
        ucc_lib_target = os.path.join(output_directory, ta_name, "lib")
        install_python_libraries(source, ucc_lib_target, python_binary_name)
        logger.info(
            f"Installed add-on requirements into {ucc_lib_target} from {source}"
        )

    ignore_list = _get_ignore_list(
        ta_name,
        os.path.abspath(os.path.join(source, os.pardir, ".uccignore")),
        output_directory,
    )
    _remove_listed_files(ignore_list)
    if ignore_list:
        logger.info(f"Removed {ignore_list} files")
    utils.recursive_overwrite(source, os.path.join(output_directory, ta_name))
    logger.info("Copied package directory")

    default_meta_conf_path = os.path.join(
        output_directory, ta_name, "metadata", meta_conf_lib.DEFAULT_META_FILE_NAME
    )
    if not os.path.exists(default_meta_conf_path):
        os.makedirs(os.path.join(output_directory, ta_name, "metadata"), exist_ok=True)
        meta_conf = meta_conf_lib.MetaConf()
        meta_conf.create_default()
        meta_conf.write(default_meta_conf_path)
        logger.info(
            f"Created default {meta_conf_lib.DEFAULT_META_FILE_NAME} file in the output folder"
        )

    with open(os.path.join(output_directory, ta_name, "VERSION"), "w") as version_file:
        version_file.write(addon_version)
        version_file.write("\n")
        version_file.write(addon_version)
        logger.info("Updated VERSION file")

    app_manifest.update_addon_version(addon_version)
    output_manifest_path = os.path.abspath(
        os.path.join(output_directory, ta_name, app_manifest_lib.APP_MANIFEST_FILE_NAME)
    )
    with open(output_manifest_path, "w") as manifest_file:
        manifest_file.write(str(app_manifest))
        logger.info(
            f"Updated {app_manifest_lib.APP_MANIFEST_FILE_NAME} file in the output folder"
        )

    app_conf = app_conf_lib.AppConf()
    output_app_conf_path = os.path.join(
        output_directory, ta_name, "default", app_conf_lib.APP_CONF_FILE_NAME
    )
    app_conf.read(output_app_conf_path)
    should_be_visible = True if global_config else False
    if global_config and global_config.meta.get("checkForUpdates") is False:
        check_for_updates = False
    else:
        check_for_updates = True
    app_conf.update(
        addon_version,
        app_manifest,
        conf_file_names,
        should_be_visible,
        check_for_updates,
    )
    app_conf.write(output_app_conf_path)
    logger.info(f"Updated {app_conf_lib.APP_CONF_FILE_NAME} file in the output folder")

    license_dir = os.path.abspath(os.path.join(source, os.pardir, "LICENSES"))
    if os.path.exists(license_dir):
        logger.info("Copy LICENSES directory")
        utils.recursive_overwrite(
            license_dir, os.path.join(output_directory, ta_name, "LICENSES")
        )

    if os.path.exists(
        os.path.abspath(os.path.join(source, os.pardir, "additional_packaging.py"))
    ):
        sys.path.insert(0, os.path.abspath(os.path.join(source, os.pardir)))
        from additional_packaging import additional_packaging

        additional_packaging(ta_name)

    if global_config:
        logger.info("Generating OpenAPI file")
        open_api_object = ucc_to_oas.transform(global_config, app_manifest)
        open_api = OpenAPI(open_api_object.json)

        output_openapi_folder = os.path.abspath(
            os.path.join(output_directory, ta_name, "static")
        )
        output_openapi_path = os.path.join(output_openapi_folder, "openapi.json")
        if not os.path.isdir(output_openapi_folder):
            os.makedirs(os.path.join(output_openapi_folder))
            logger.info(f"Creating {output_openapi_folder} folder")
        with open(output_openapi_path, "w") as openapi_file:
            json.dump(open_api.raw_element, openapi_file, indent=4)
